/*
 * Copyright (c) 2009 Silvia Ramis, Marta Ramis
 * Copyright (c) 2016 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This code was originally written by Silvia Ramis and Marta Ramis
 * for their Final Project on Computer Sciences Engineering (2009)
 * under the supervision of J.L. Lisani
 * Some parts of the code were written by J.L. Lisani
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/****
This code is associated with the following publication

J.L. Lisani, S. Ramis, 
A contrario detection of faces with a short cascade of classifiers
Image Processing On Line (IPOL), 2019
****/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <dirent.h>
#include <time.h>

#include <vector>
#include <algorithm>
using namespace std;


#include "io_png/io_png.h"
#include "io_RGB.h"
#include "tools_lib.h"
#include "inversegaussiancdf.h"
#include "randmt.h"
#include "lib_features.h"
#include "lib_facedet.h"


//Fixed parameters

//standard image width and height for faces 24x24
#define wSTD 24
#define hSTD 24
//minimum percentage of common area to group together two detections
#define AREACOMMON 0.5 //50%


//Global variables
//minimum and maximum size of detected sub-windows
int sizeMin, sizeMax;
//zoom factor for rescaling sub-windows to standard size 24 x 24
float zStep; 


//Detection result data
struct detection_result {
    int id; //ID of the result
    int x1, y1, x2, y2; //position in image
    float vdet; //detection value 
    struct detection_result *next;
    //Auxiliary variables:
    float zoom;//zoom factor that rescales subwindow 
               //of size (x2-x1+1) x (y2-y1+1)
               //to standard-size 24 x 24
    int x0, y0;//position of top-left corner of sub-window in image
               //after rescaling with factor 'zoom' 
};

//List of all detection results
struct detection_results {
    int ndetections;
    struct detection_result **tabdet;
    struct detection_result *first, *last;
};

//get intensity channel and size of input image
unsigned char *read_image_data(const char *namein, int &w, int &h)
{
    unsigned char *I, *R, *G, *B;

    size_t ww, hh;
    unsigned char *img; //image data
    /* read the PNG input image */
    img = io_png_read_u8_rgb(namein, &ww, &hh);
    w=(int) ww;
    h=(int) hh;
    input2RGB(img, &R, &G, &B, w*h);
    I=RGBtoI(R, G, B, w*h);
    
    free(img);
    delete[] R;
    delete[] G;
    delete[] B;
    
    return I;
}



//This function implements lines 1 to 3 of Algorithm 3
//as described in the associated IPOL publication

//compute set of zoom factors that permit to normalize subwindows 
//of sizes s x s (with s in range [sizeMin, sizeMax]) to 
//standard size 24 x 24
vector<float> zoom_factors()
{
    int sizeref=wSTD;
    vector<float> zooms;
    float z_prev=0;
    for (int s=sizeMin; s <= sizeMax; ) {
        //round zoom factor to nearest float with 2 decimal precision
        float z=roundf( ((float) sizeref/(float) s) * 100) / 100;
        if ((z > 0) && (z != z_prev)) { //check that new zoom factor is different from
                             //previous one
            zooms.push_back(z);
            z_prev=z;
        }
        if (zStep > 1) s*=zStep;
        else s++;
    }

    return zooms;
}




//check features of subimage
//if usemirror=TRUE then check is made on mirror version of input
//but un-mirrored coordinates must be stored
void check_subimage(unsigned char *Izoom, unsigned char *Istd, 
                    float *If, float *Ii,
                    int x0, int y0, int wzoom, 
                    float zoom, vector<classifier_item> &cdata,
                    vector<struct detection_result> &dres,
                    unsigned char usemirror)
{
    float vdet;

    getsubimage(Izoom, Istd, x0, y0, wzoom, wSTD, hSTD);    
        
    //compute detection value for subimage
    if (check_features_imageStandard(Istd, wSTD, hSTD, If, Ii, cdata, vdet)) {

        struct detection_result dr;
        dr.vdet=vdet;
        dr.zoom=zoom;
        if (usemirror) { // 'un-mirror' x-coordinate   xm0=wzoom-x0-wSTD;
            x0=wzoom-x0-wSTD; 
        }
        dr.x0=x0;
        dr.y0=y0;
        dr.x1=(int) (((float) x0)/zoom + 0.5f);
        dr.y1=(int) (((float) y0)/zoom + 0.5f);
        dr.x2=(int) (((float) (x0+wSTD-1))/zoom + 0.5f);
        dr.y2=(int) (((float) (y0+hSTD-1))/zoom + 0.5f);
        
        dres.push_back(dr);
    } 


}

//apply classifier on subwindows already checked with another classifier
//at a given scale (zoom factor)
//if usemirror=TRUE then check subimage from mirror version of input
vector<struct detection_result> check_featuresimage(unsigned char *I, 
                                                    int w, int h,
                                                    unsigned char *Istd, 
                                                    float *If, float *Ii,
                                                    float zoom, 
                                                    vector<classifier_item> &cdata,
                                                    vector<struct detection_result> &in,
                                                    unsigned char usemirror)
{
    int wzoom, hzoom;
    unsigned char *Izoom, *Imirror;
    int x0, y0, xm0;
    unsigned char isface;
    float vdet;
    vector<struct detection_result> dres;

    //rescale input image
    if (zoom == 1) {
        wzoom=w;
        hzoom=h;
        Izoom=I;
    } else {
        Izoom=zoom_image(I, w, h, zoom, wzoom, hzoom);
    }
    
    //compute mirror image if usemirror=TRUE
    if (!usemirror) {
        Imirror=Izoom;
    } else {
        Imirror=mirror_image(Izoom, wzoom, hzoom);
    }


    if ((wzoom >= wSTD) && (hzoom >= hSTD)) {

        if (in.empty()) {
          //check whole image
          for (y0=0; y0+hSTD-1 < hzoom; y0++) {
              for (x0=0; x0+wSTD-1 < wzoom; x0++) {
                  xm0=x0;
                  if (usemirror) xm0=wzoom-x0-wSTD; //top-left corner in mirror image
                  check_subimage(Imirror, Istd, If, Ii, xm0, y0, 
                                 wzoom, zoom, cdata, dres, usemirror);
              }
          }
            
        } else {
            //check only input sub-windows
            for (int k=0; k < in.size(); k++) {
                if (in[k].zoom == zoom) { //check if selected subwindow has been 
                                           //obtained at the current scale
                    x0=in[k].x0;
                    y0=in[k].y0;
                    xm0=x0;
                    if (usemirror) xm0=wzoom-x0-wSTD; //top-left corner in mirror image
                    check_subimage(Imirror, Istd, If, Ii, xm0, y0, 
                                   wzoom, zoom, cdata, dres, usemirror);
                }
            }
        }

    }

    if (Imirror != Izoom) delete[] Imirror;
    if (zoom != 1) delete[] Izoom;
    
    return dres;

}


//apply classifier on subwindows already checked with another classifier
//at different scales (zoom factors)
vector<struct detection_result> applyclassifier(unsigned char *I, 
                                                int w, int h,
                                                const char *nameclass,
                                                vector<struct detection_result> &in,
                                                unsigned char usemirror)
{
    //read classifier data from file
    vector<classifier_item> cdata = read_classifier_data(nameclass);

    //set of zoom factors that permit to detect faces of size s x s
    //with s in range [sizeMin, sizeMax]
    vector<float> zooms=zoom_factors();
    int nzoom = zooms.size();
       
    //auxiliary variable
    //indicates which zoom factors need to be used in the computations
    unsigned char *applyzoom = new unsigned char[nzoom];
    memset(applyzoom, 0, nzoom);
    if (!in.empty()) {
        //classifier applied on selected image subwindows: 
        //check if zoom factor is used
        for (int izoom=0; izoom < nzoom; izoom++) {
            for (int k=0; (k < in.size()) && (applyzoom[izoom] == 0); k++) {
                if (in[k].zoom == zooms[izoom]) { //check if selected subwindow has been 
                                                  //obtained with the current zoom factor
                    applyzoom[izoom] = 1;
                }
            }
        }
    } else {
        //classifier applied on all image subwindows: use all zoom factors
        memset(applyzoom, 1, nzoom);
    }

      
    //detection results for each zoom factor
    vector<struct detection_result> dres[nzoom];
    
    //parallel computation for each zoom factor
#pragma omp parallel for schedule(dynamic)
    for (int izoom=0; izoom < nzoom; izoom++) {
        if (applyzoom[izoom] != 0) {
            //auxiliary variables passed as parameter
            //allocated only once per zoom factor
            //this type of allocation permits parallelization of the for loop with OMP
            unsigned char *Istd=new unsigned char[wSTD*hSTD];
            float *If=new float[wSTD*hSTD];
            float *Ii=new float[wSTD*hSTD];
            
            dres[izoom] = check_featuresimage(I, w, h, Istd, If, Ii, 
                                              zooms[izoom], cdata, in, usemirror);
                                                                                            
            //delete auxiliary variables
            delete[] If;
            delete[] Ii;
            delete[] Istd;
        }    
    }
    
    //combine all detection results into a single data structure
    vector<struct detection_result> dresT;
    for (int izoom=0; izoom < nzoom; izoom++) {
        dresT.insert( dresT.end(), dres[izoom].begin(), dres[izoom].end() );
    }
      
    delete[] applyzoom;
    
    return dresT;
}




//sort data from highest to lowest vdet
bool sort_decreasing_vdet(const detection_result &a, const detection_result &b)
{
    return a.vdet > b.vdet;
}


//This function implements Algorithm 6,
//as described in the associated IPOL publication

//Organize detections in a tree structure, then prune the tree.
//For each branch of the tree keep detection with highest detection value
//Group detections based on following criterium:
//A and B are grouped is S(A, B)=Intersection(A, B)/Union(A, B) > threshold 
//(default value for threshold: 0.5)
vector<struct detection_result> detections_prune(vector<struct detection_result> in)
{
    vector<struct detection_result> out;


    if (in.empty()) return out;

    //sort detections
    sort(in.begin(), in.end(), sort_decreasing_vdet);
    
    //auxiliary data structure: keep only detections with flag=1
    unsigned char *active=new unsigned char[in.size()];
    memset(active, 1, in.size());
    
    //check superpositions
    float prccommon;
    for (int n=in.size()-1; n >= 0; n--) {
        for (int m=n-1; (m >= 0) && active[n]; m--) {
            prccommon=check_superposition_rectangles(in[n].x1, in[n].y1,
                                                     in[n].x2, in[n].y2,
                                                     in[m].x1, in[m].y1,
                                                     in[m].x2, in[m].y2);

            if (prccommon > AREACOMMON) { //same group of detections, 
                                          //keep the one with highest detection value
                active[n]=0;
            }

        }
    }

    //output detections
    for (int n=0; n < in.size(); n++) 
        if (active[n]) out.push_back(in[n]);
       
    //free auxiliary data structure
    delete[] active;

    return out;
}


//display detections as green squares on original input image
void display_detections(vector<struct detection_result> res, 
                        const char *namein, const char *nameout)
{
    //read input image
    size_t ww, hh;
    int w, h; //input image dimensions
    unsigned char *img; //image data
    unsigned char *R, *G, *B; //image data organized in R, G and B channels
    /* read the PNG input image */
    img = io_png_read_u8_rgb(namein, &ww, &hh);
    w=(int) ww;
    h=(int) hh;
    input2RGB(img, &R, &G, &B, w*h);

    int x1, y1, x2, y2;
    for (int n=0; n < res.size(); n++) {
        x1=res[n].x1;
        y1=res[n].y1;
        x2=res[n].x2;
        y2=res[n].y2;

        //display in green
        drawRectangleGrey(R, w, h, x1, y1, x2, y2, 0);
        drawRectangleGrey(G, w, h, x1, y1, x2, y2, 255);
        drawRectangleGrey(B, w, h, x1, y1, x2, y2, 0);

        //enhance display (thicker lines)
        if ((x1-1 >= 0) && (x2+1 < w) && (y1-1 >= 0) && (y2+1 < h)) {
            drawRectangleGrey(R, w, h, x1-1, y1-1, x2+1, y2+1, 0);
            drawRectangleGrey(G, w, h, x1-1, y1-1, x2+1, y2+1, 255);
            drawRectangleGrey(B, w, h, x1-1, y1-1, x2+1, y2+1, 0);
        }
        if ((x1-2 >= 0) && (x2+2 < w) && (y1-2 >= 0) && (y2+2 < h)) {
            drawRectangleGrey(R, w, h, x1-2, y1-2, x2+2, y2+2, 0);
            drawRectangleGrey(G, w, h, x1-2, y1-2, x2+2, y2+2, 255);
            drawRectangleGrey(B, w, h, x1-2, y1-2, x2+2, y2+2, 0);
        }

    }

    printf("%i detections displayed\n", (int) res.size());


    //save output image
    RGB2output(R, G, B, img, w*h);
    io_png_write_u8(nameout, img, w, h, 3); //3 channels

    free(img);
    delete[] R;
    delete[] G;
    delete[] B;

}



//This function implements 'SampleSubWindows', called from Algorithm 2,
//as described in the associated IPOL publication

//randomly select nrandom detections among all the input detections 
vector<struct detection_result> random_detections(vector<struct detection_result> &in, 
                                                  int nrandom)
{
    vector<struct detection_result> out;

    if (in.size() <= nrandom) return out;

    //select random patches, without reposition
    init_randmt_auto();
    unsigned char *repeatedpatch=new unsigned char[in.size()];
    memset(repeatedpatch, 0, in.size());
    int np=0;
    while (np < nrandom) {
        double a=rand_unif(); //random number between 0 and 1
        int idpatch=(int) (a*(double) (in.size()-1)); //random integer number 
                                                      //between 0 and ndata-1
        //check that detection has not already been used
        if (!repeatedpatch[idpatch]) {
            out.push_back(in[idpatch]);
            repeatedpatch[idpatch]=1; //mark detection as used
            np++;
        }
    }
    delete[] repeatedpatch;

    return out;
}



//This function implements 'GetMeanStandardDev', called from Algorithm 2,
//as described in the associated IPOL publication

//compute mean and standard of detection values stored in 'res' data structure
void detections_stats(vector<struct detection_result> res, 
                      double &vmean, double &vstd)
{
    //sort input data
    sort(res.begin(), res.end(), sort_decreasing_vdet);

    //mean of values
    vmean=0;
    for (int n=0; n < res.size(); n++) vmean+=res[n].vdet;
    vmean/=(double) res.size();

    //compute std
    vstd=0;
    for (int n=0; n < res.size(); n++) {
        vstd+=(res[n].vdet-vmean)*(res[n].vdet-vmean);
    }
    vstd/=(double) (res.size()-1); //sample variance
    vstd=sqrt(vstd);

    printf("Stats:\n");
    printf("%i values   vmean=%2.2f  vstd=%2.2f\n", (int) res.size(), vmean, vstd);

}

//keep only detections whose detection value is above a threshold
vector<struct detection_result> detections_threshold(vector<struct detection_result> &in, 
                                                     double threshold)
{
    vector<struct detection_result> out;


    if (in.empty()) return out;

    for (int n=0; n < in.size(); n++) 
        if (in[n].vdet > threshold) 
            out.push_back(in[n]);
            
    return out;
}


//This function implements Algorithm 4 and part of Algorithm 2,
//as described in the associated IPOL publication

//Face detection function
//Input arguments:
//  - name of input image
//  - name of first classifier
//  - name of second classifier
//  - name of third classifier
//  - name of fourth classifier
//  - minimum size of detected faces
//  - maximum size of detected faces
//  - size increase factor from minimum to maximum size
//  - percentage of detections kept by 1st classifier
//  - percentage of detections kept by 2nd classifier
//  - maximum number of false positives allowed by 3rd classifier
//  - maximum number of false positives allowed by 4th classifier
//    = maximum number of false positives at the detector's output
void detectfaces(const char *namein,
                  const char *nameclass1,
                  const char *nameclass2,
                  const char *nameclass3,
                  const char *nameclass4,
                  const char *nameout,
                  int minSize,
                  int maxSize,
                  float stepZoom,
                  float prc1,
                  float prc2,
                  float nfa1,
                  float nfa2)
{
    unsigned char *I;
    int w, h;
    I = read_image_data(namein, w, h);    
    
    //set global variables
    sizeMin = minSize;
    sizeMax = maxSize;
    zStep = stepZoom;
    
    printf("Maximum number of false positives: %2.2f\n", nfa2);
    printf("Minimum size of the detected faces: %i\n", sizeMin);
    printf("Maximum size of the detected faces: %i\n", sizeMax);
    printf("Size increase factor: %2.2f\n", zStep);
    
   
    //compute detection value for all possible sub-windows using 
    //5-features classifier
    vector<struct detection_result> res0; //auxiliary variable, empty set
    vector<struct detection_result> res1;
    res1=applyclassifier(I, w, h, nameclass1, res0, 0); 
    if (res1.size() == 0) {
        //this may happen when all sub-windows are flat 
        //(i.e. have low variance)
        printf("0 raw detections\n"); 
        delete[] I;
        return;
    }
    
    //compute detection thresholds
    double threshold1, threshold2, threshold3, threshold4;
    
    //compute detection threshold for the 5-features classifier:
    //threshold1 = (1-prc1)-quantile of detection values
    //sort sub-windows by detection value
    sort(res1.begin(), res1.end(), sort_decreasing_vdet);
    //compute (1-prc1)-quantile
    if ((int) roundf(prc1*(res1.size()-1)) < 0)  { 
        //this may happen when almost all sub-windows are flat 
        //(i.e. have low variance)
        printf("0 raw detections\n"); 
        delete[] I;
        return;
    }
    threshold1=res1[(int) roundf(prc1*(res1.size()-1))].vdet;
    printf("threshold1=%2.2f \n", threshold1);
    
    //pick randomly npatches sub-windows and use them to compute stats for the
    //rest of classifiers
    //printf("Thresholds from random detections\n");
    float ppatches=0.01f; //percentage of randomly chosen sub-windows
    int npatches=(int) (ppatches*(float) (res1.size()));
    if (npatches == 0) {
        //this may happen when almost all sub-windows are flat 
        //(i.e. have low variance)
        printf("0 raw detections\n"); 
        delete[] I;
        return;    
    }
    printf("Pick %i random patches\n", npatches);
    //random selection from all the possible sub-windows
    vector<struct detection_result> resR1=random_detections(res1, npatches);
    
    //compute detection threshold for the 10-features classifier:
    //threshold2 = (1-prc2)-quantile of detection values of randomly chosen
    //sub-windows
    //sort sub-windows by detection value
    vector<struct detection_result> resR2;
    resR2=applyclassifier(I, w, h, nameclass2, resR1, 0);
    sort(resR2.begin(), resR2.end(), sort_decreasing_vdet);
    //compute (1-prc2)-quantile
    if ((int) roundf(prc2*(resR2.size()-1)) < 0)  { 
        //this may happen when almost all sub-windows are flat 
        //(i.e. have low variance)
        printf("0 raw detections\n"); 
        delete[] I;
        return;
    }
    threshold2=resR2[(int) roundf(prc2*(resR2.size()-1))].vdet;
    printf("threshold2=%2.2f \n", threshold2);

    //compute detection threshold for the 80-features classifier
    //using Equation (10)
    vector<struct detection_result> resR3;
    resR3=applyclassifier(I, w, h, nameclass3, resR1, 0); 
    double vmean3, vstd3;
    detections_stats(resR3, vmean3, vstd3);
    double sensitivity3=invcumZabove((double) nfa1/(double) res1.size());
    //printf("sensitivity3=%2.2f  (NFP3=%2.2f)\n", sensitivity3, nfa1);
    threshold3=vmean3+sensitivity3*vstd3;
    printf("threshold3=%2.2f \n", threshold3);


    //compute detection threshold for the 200-features classifier
    //using Equation (10)
    vector<struct detection_result> resR4;
    resR4=applyclassifier(I, w, h, nameclass4, resR1, 0); 
    double vmean4, vstd4;
    detections_stats(resR4, vmean4, vstd4);
    double sensitivity4=invcumZabove((double) nfa2/(double) res1.size());
    //printf("sensitivity4=%2.2f  (NFP4=%2.2f)\n", sensitivity4, nfa2);
    threshold4=vmean4+sensitivity4*vstd4;
    printf("threshold4=%2.2f \n", threshold4);



    //apply detection thresholds
    vector<struct detection_result> resT1;
    vector<struct detection_result> resT2;
    vector<struct detection_result> resT3;
    vector<struct detection_result> resT4;

    vector<struct detection_result> res2;
    vector<struct detection_result> res3;
    vector<struct detection_result> res4;

    //get detections from 5-features classifier 
    //with detection value above threshold1
    resT1=detections_threshold(res1, threshold1);
    if (!resT1.empty()) {
        printf("Detector 1: %i detections (%i initial)\n", (int) resT1.size(), (int) res1.size());
        
        //apply 10-features classifier to selected sub-windows
        res2=applyclassifier(I, w, h, nameclass2, resT1, 0); 
        if (!res2.empty()) {
        
            //get detections from 10-features classifier 
            //with detection value above threshold2                
            resT2=detections_threshold(res2, threshold2);
            if (!resT2.empty()) {
                printf("Detector 2: %i detections (%i initial)\n", (int) resT2.size(), (int) res2.size());
                
                //apply 80-features classifier to selected sub-windows                
                res3=applyclassifier(I, w, h, nameclass3, resT2, 0); 
                if (!res3.empty()) {
                
                    //get detections from 80-features classifier 
                    //with detection value above threshold3                
                    resT3=detections_threshold(res3, threshold3);
                    if (!resT3.empty()) {
                        printf("Detector 3: %i detections (%i initial)\n", (int) resT3.size(), (int) res3.size());
                        
                        //apply 200-features classifier to selected sub-windows    
                        res4=applyclassifier(I, w, h, nameclass4, resT3, 0); 
                        if (!res4.empty()) {
                        
                            //get detections from 80-features classifier 
                            //with detection value above threshold3                
                            resT4=detections_threshold(res4, threshold4);
                            
                            if (!resT4.empty()) {
                                printf("Detector 4: %i detections (%i initial)\n", (int) resT4.size(), (int) res4.size());
                            } else printf("Detector 4: 0 detections (%i initial)\n", (int) res4.size());
                        } else printf("Detector 4: 0 initial detections\n");
                    } else printf("Detector 3: 0 detections (%i initial)\n", (int) res3.size());
                } else printf("Detector 3: 0 initial detections\n");
            } else printf("Detector 2: 0 detections (%i initial)\n", (int) res2.size());
        } else printf("Detector 2: 0 initial detections\n");
    } else printf("Detector 1: 0 detections (%i initial)\n", (int) res1.size());


    //group together overlapping detections
    if (!resT4.empty()) {
        printf("Number of raw detections=%i\n", (int) resT4.size());
        char namedataout[120];
                     
	    //keep only 'stable' detections (mirror filter)
        printf("Mirror filter\n");
        //compute detections in mirror image using 200-features classifier
        vector<struct detection_result> resM;
        resM=applyclassifier(I, w, h, nameclass4, resT4, 1); 
        //keep detections with detection value above threshold4       
        vector<struct detection_result> resTM;         
        resTM=detections_threshold(resM, threshold4);
        
        //group together overlapping detections
        if (!resTM.empty()) {
            printf("Number of mirror detections=%i\n", (int) resTM.size());
            char namedataout[120];
                 
            vector<struct detection_result> resAux=detections_prune(resTM);
            
            printf("Final number of detections=%i\n", (int) resAux.size());

            //display detections
            strcpy(namedataout, nameout);
            strcat(namedataout, ".png");
            display_detections(resAux, namein, (const char *) namedataout);            
            
            //write results to .txt file
            strcpy(namedataout, nameout);
            strcat(namedataout, ".txt");
            FILE *f=fopen(namedataout, "w");
            fprintf(f, "%i %i\n", w, h);
            int x1, y1, x2, y2;
            for (int n=0; n < resAux.size(); n++) {
                x1=resAux[n].x1;
                y1=resAux[n].y1;
                x2=resAux[n].x2;
                y2=resAux[n].y2;
                fprintf(f, "%i %i %i %i\n", x1, y1, x2, y2);
            }
            fclose(f);
            
            //create file of detection statistics
            //(needed if we want to compute NFP values 
            //associated to each detection)
            //NFP(detection) = P(detection value) * Ntestedsubwindows
            strcpy(namedataout, nameout);
            strcat(namedataout, "_det_stats.txt");
            f=fopen(namedataout, "w");
            //number of tested subwindows
            fprintf(f, "%i\n", (int) res1.size());
            //mean and std of detection values distribution for 200-features
            //classifier
            fprintf(f, "%2.6f %2.6f\n", vmean4, vstd4);
            //detection values of final detections
            for (int n=0; n < resAux.size(); n++) 
                fprintf(f, "%2.6f\n", resAux[n].vdet);
            fclose(f);
            
            
        } else printf("0 mirror detections\n");       
        
    } else printf("0 raw detections\n");     

    delete[] I;
    

}


