/*
 * Copyright (c) 2009 Silvia Ramis, Marta Ramis
 * Copyright (c) 2016 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * Copyright (c) 2019 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This code was originally written by Silvia Ramis and Marta Ramis
 * for their Final Project on Computer Sciences Engineering (2009)
 * under the supervision of J.L. Lisani
 * Some parts of the code were written by J.L. Lisani
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
/****
This code is associated with the following publication

J.L. Lisani, S. Ramis, 
A contrario detection of faces with a short cascade of classifiers
Image Processing On Line (IPOL), 2019

 
 
Additional reference:
 
[ViolaJones2001] P. Viola and M. Jones, Rapid object detection using a boosted 
cascade of simple features, 
in Proceedings of the 2001 IEEE Computer Society Conference on Computer Vision 
and Pattern Recognition, CVPR, vol. 1, 2001, pp. 511–518, 
https://doi.org/10.1109/CVPR.2001.990517.

****/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <dirent.h>
#include <time.h>

#include "io_png/io_png.h"
#include "io_RGB.h"
#include "randmt.h"
#include "lib_features.h"

//Global variable
#define FLAT_IMAGE_VAR 400 //minimum std of subimages=20


//This function implements 'NormalizeVariance', called from Algorithm 3,
//as described in the associated IPOL publication

double normalize_variance(unsigned char *I, float *If, int w, int h)
{
    double VNorm=2500;//sdevnorm=50, normalized variance
    
    //compute variance
    double v, var=0, med=0;
    for (int n=0; n < w*h; n++) {
        v=(double) I[n];
        med+=v;
        var+=v*v;
    }
    med/=(double) (w*h);
    var=var/(double) (w*h)-med*med;
    
    //compute normalization factor
    float a;
    if (var != 0) a=(float) sqrt(VNorm/var);
    else a=1.0f;
    
    //normalize image values
    for (int n=0; n < w*h; n++) If[n]=a*((float) I[n]);

    return var;
}

//compute integral image, as described in [ViolaJones2001] paper
void Integralf(float *in, float *out, int width, int height)
{
    float *s=new float[width*height];
    int i=0, j=0, a=0, b=0;
    for (j=0; j<height; j++) {
        for(i=0; i<width; i++) {
            if (j==0) { 
                s[i+(j*width)] = in[i+(j*width)];
            } else {
                s[i+(j*width)] = s[i+((j-1)*width)] + in[i+(j*width)];
            }
            if (i==0) { 
                out[i+(j*width)]= s[i+(j*width)];
            } else {
                out[i+(j*width)]= out[(i-1)+(j*width)] + s[i+(j*width)];
            }
        }
    }
    delete[] s;
}

//compute sum of pixel values in a rectangle, using integral image
float recIntegral(float *Ii, int w, int h, int x1, int y1, int x2, int y2)
{
    float s1, s2, s3, s4, sum;

    s1=Ii[x2+y2*w];
    s2=(x1 >= 1)?(Ii[x1-1+y2*w]):(0);
    s3=(y1 >= 1)?(Ii[x2+(y1-1)*w]):(0);
    s4=((x1 >= 1) && (y1 >= 1))?(Ii[(x1-1)+(y1-1)*w]):(0);
    sum=s1-s2-s3+s4;

    return sum;
}


//compute value of Haar-like feature of type 
// WWWWBBBB 
// WWWWBBBB 
// WWWWBBBB 
// WWWWBBBB 
//(see Figure 1a in associated IPOL paper)
float image_feature2V(float *Ii, int w, int h, int x1, int y1, int x2, int y2)
{
    float sum1, sum2;
    int ww, hh, xm;

    ww=x2-x1+1;
    hh=y2-y1+1;
    xm=x1+ww/2-1;
    sum1=recIntegral(Ii, w, h, x1, y1, xm, y2);
    sum2=recIntegral(Ii, w, h, xm+1, y1, x2, y2);

    return (sum1-sum2);
}

//compute value of Haar-like feature of type 
// BBBBBBBB 
// BBBBBBBB 
// WWWWWWWW 
// WWWWWWWW 
//(see Figure 1c in associated IPOL paper)
float image_feature2H(float *Ii, int w, int h, int x1, int y1, int x2, int y2)
{
    float sum1, sum2;
    int ww, hh, ym;

    ww=x2-x1+1;
    hh=y2-y1+1;
    ym=y1+hh/2-1;
    sum1=recIntegral(Ii, w, h, x1, y1, x2, ym);
    sum2=recIntegral(Ii, w, h, x1, ym+1, x2, y2);

    return (sum1-sum2);
}


//compute value of Haar-like feature of type 
// WWWWBBBBWWWW 
// WWWWBBBBWWWW 
// WWWWBBBBWWWW 
// WWWWBBBBWWWW 
//(see Figure 1b in associated IPOL paper)
float image_feature3V(float *Ii, int w, int h, int x1, int y1, int x2, int y2)
{
    float sum1, sum2, sum3;
    int ww, hh, xm1, xm2;

    ww=x2-x1+1;
    hh=y2-y1+1;
    xm1=x1+ww/3-1;
    xm2=x1+2*ww/3-1;
    sum1=recIntegral(Ii, w, h, x1, y1, xm1, y2);
    sum2=recIntegral(Ii, w, h, xm1+1, y1, xm2, y2);
    sum3=recIntegral(Ii, w, h, xm2+1, y1, x2, y2);

    return (sum1-sum2+sum3);
}


//compute value of Haar-like feature of type 
// WWWWWWWW 
// WWWWWWWW 
// BBBBBBBB 
// BBBBBBBB 
// WWWWWWWW 
// WWWWWWWW 
//(see Figure 1d in associated IPOL paper)
float image_feature3H(float *Ii, int w, int h, int x1, int y1, int x2, int y2)
{
    float sum1, sum2, sum3;
    int ww, hh, ym1, ym2;

    ww=x2-x1+1;
    hh=y2-y1+1;
    ym1=y1+hh/3-1;
    ym2=y1+2*hh/3-1;
    sum1=recIntegral(Ii, w, h, x1, y1, x2, ym1);
    sum2=recIntegral(Ii, w, h, x1, ym1+1, x2, ym2);
    sum3=recIntegral(Ii, w, h, x1, ym2+1, x2, y2);

    return (sum1-sum2+sum3);
}


//compute value of Haar-like feature of type 
// WWWWBBBB 
// WWWWBBBB 
// BBBBWWWW
// BBBBWWWW 
//(see Figure 1e in associated IPOL paper)
float image_feature4(float *Ii, int w, int h, int x1, int y1, int x2, int y2)
{
    float sum1, sum2, sum3, sum4;
    int ww, hh, xm, ym;

    ww=x2-x1+1;
    hh=y2-y1+1;
    xm=x1+ww/2-1;
    ym=y1+hh/2-1;
    sum1=recIntegral(Ii, w, h, x1, y1, xm, ym);
    sum2=recIntegral(Ii, w, h, xm+1, y1, x2, ym);
    sum3=recIntegral(Ii, w, h, x1, ym+1, xm, y2);
    sum4=recIntegral(Ii, w, h, xm+1, ym+1, x2, y2);

    return (sum1-sum2-sum3+sum4);
}


//Read classifier data:
//* number of features masks
//* for each mask:
//  - position in 24x24 image
//  - type
//  - classification threshold
//  - threshold sign
//  - learned alpha value
vector<classifier_item> read_classifier_data(const char *name)
{
    vector<classifier_item> cdata;
    FILE *f=fopen(name, "r");
    if (!f) {
        printf("File %s not found\n", name);
        return cdata;
    }

    int ndata;
    int aux = fscanf(f, "%i", &ndata);
    for (int n=0; n < ndata; n++) {
        struct classifier_item item;
        int aux = fscanf(f, "%i %i %i %i %i %f %i %f",
                  &item.x1, &item.y1, &item.x2, &item.y2,
                  &item.type, &item.vth, &item.polth,
                  &item.alpha);
        cdata.push_back(item);
    }
    fclose(f);
    
    return cdata;
}


//Compute output value of the classifier when applied on a subimage
//standard subimage size: same size as training samples (default: 24 x 24)
//return 0 if subimage is flat
//else return 1 and the output value of the classifier 
unsigned char check_features_imageStandard(unsigned char *I, int w, int h, 
                                           float *If, float *Ii,
                                           vector<classifier_item> cdata, 
                                           float &vdet)
{
    //normalize variance (result in If)
    double var=normalize_variance(I, If, w, h);
    if (var < FLAT_IMAGE_VAR) return 0; //flat image

    //compute auxiliary integral image (result in Ii)
    Integralf(If, Ii, w, h);
    
    //check features
    double sum, alphasum=0;
    unsigned char type, hm;
    int x1, y1, x2, y2;
    for (int m=0; m < cdata.size(); m++) {
        type=(unsigned char) cdata[m].type;
        x1=cdata[m].x1;
        y1=cdata[m].y1;
        x2=cdata[m].x2;
        y2=cdata[m].y2;
        switch (type) {
            case 1: {
                sum=image_feature2V(Ii, w, h, x1, y1, x2, y2);
                break;
            }
            case 2: {
                sum=image_feature2H(Ii, w, h, x1, y1, x2, y2);
                break;
            }
            case 3: {
                sum=image_feature3V(Ii, w, h, x1, y1, x2, y2);
                break;
            }
            case 4: {
                sum=image_feature3H(Ii, w, h, x1, y1, x2, y2);
                break;
            }
            case 5: {
                sum=image_feature4(Ii, w, h, x1, y1, x2, y2);
                break;
            }
            default:
            printf("Type Error!!\n");
        }
        hm=0; //hm: result of weak classifier for feature m
        if ((sum <= cdata[m].vth) && (cdata[m].polth == 1)) hm=1;
        if ((sum > cdata[m].vth) && (cdata[m].polth == -1)) hm=1;
        
        //Output of strong classifier: Equation XX in associated paper
        alphasum+=((double) hm)*cdata[m].alpha;  
        
    }
    
    vdet=alphasum;
    
    return 1;
    
}

