/*
 * Copyright (c) 2009 Silvia Ramis, Marta Ramis
 * Copyright (c) 2013 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This code was originally written by Silvia Ramis and Marta Ramis
 * for their Final Project on Computer Sciences Engineering (2009)
 * under the supervision of J.L. Lisani
 * Some parts of the code were written by J.L. Lisani
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 
/****
This code is associated with the following publication

J.L. Lisani, S. Ramis, 
A contrario detection of faces with a short cascade of classifiers
Image Processing On Line (IPOL), 2019
****/



#include <stdio.h>
#include <string.h>
#include <math.h>

//new coordinate value after zoom
int zoomed(int i, float z)
{
    return ((float) (((float) i)*z));
}



//zoom >1 
//use bilinear interpolation
void zoom_in_image(unsigned char *in, unsigned char *out,
                   float zoom, int width, int height,
                   int new_width, int new_height)
{
    int i, j, zi, zj;
    int ii, jj, l1, l2, l3, l4;
    float x, y, a, b, c, d;

    for (i=0; i < new_width; i++) {
        x=((float) i)/zoom;
        ii=(int) x;
        x-=(float) ii;
        for (j=0; j < new_height; j++) {
            y=((float) j)/zoom;
            jj=(int) y;
            y-=(float) jj;
            l1=in[ii+jj*width];
            if (ii < width) l2=in[ii+1+jj*width];
            else l2=l1;
            if (jj < height) l3=in[ii+(jj+1)*width];
            else l3=l1;
            if ((ii < width) && (jj < height)) l4=in[ii+1+(jj+1)*width];
            else l4=l1;
            a=(float) (l4+l1-l2-l3);
            b=(float) (l2-l1);
            c=(float) (l3-l1);
            d=(float) l1;
            out[i+j*new_width]=(int) (a*x*y+b*x+c*y+d);
        }
    }

}


//zoom < 1 
//average + downsampling
void zoom_out_image(unsigned char *in, unsigned char *out,
                    float zoom, int width, int height,
                    int new_width, int new_height)
{
    int i, j, zi, zj, npixels, ng, ngm;

    for (zi=0; zi < new_width; zi++)
        for (zj=0; zj < new_height; zj++) {
            npixels=0;
            ng=0;
            for (i=zoomed(zi, 1/zoom); i < zoomed(zi+1, 1/zoom); i++)
                for (j=zoomed(zj, 1/zoom); j < zoomed(zj+1, 1/zoom); j++) {
                    if (i >= width) i=width-1;
                    if (j >= height) j=height-1;
                    ng+=(int) in[i+j*width];
                    npixels++;
                }
            //store average value
            ngm=(int) (((float) ng)/((float) npixels));
            out[zi+zj*new_width]=ngm;
        }
        
        
}

//This function implements 'RescaleImage', called from Algorithm 3,
//as described in the associated IPOL publication

//zoom input image
void zoom_image(unsigned char *in, unsigned char *out,
                int width, int height,
                float zoom, int new_width, int new_height)
{
    if (zoom == 1.0) 
        memcpy(out, in, width*height);

    if (zoom > 1.0) 
        zoom_in_image(in, out, zoom, width, height, new_width, new_height);

    if (zoom < 1.0) 
        zoom_out_image(in, out, zoom, width, height, new_width, new_height);

}

//zoom input image
//create output image
unsigned char *zoom_image(unsigned char *in, int width, int height,
                          float zoom, int &new_width, int &new_height)
{
    new_width=zoomed(width, zoom);
    new_height=zoomed(height, zoom);
    unsigned char *out=new unsigned char[new_width*new_height];
    
    zoom_image(in, out, width, height, zoom, new_width, new_height);

    return out;
}

//This function implements 'HorizontalFlip', called from Algorithm 5,
//as described in the associated IPOL publication

//mirror version of input image
//create output image
unsigned char *mirror_image(unsigned char *in, int width, int height)
{
    unsigned char *out=new unsigned char[width*height];
    for (int y=0; y < height; y++)
        for (int x=0; x < width; x++) 
            out[x+y*width] = in[width-1-x+y*width];
  
    return out;  
}
                          


//This function implements 'GetSubWindow', called from Algorithm 3,
//as described in the associated IPOL publication

//get subimage from I
//top-left corner: (Tx, Ty)
//bottom-right corner: (Tx+wsub-1, Ty+hsub-1)
void getsubimage(unsigned char *I, unsigned char *Isub, 
                 int Tx, int Ty, int w, int wsub, int hsub)
{
    for (int y=Ty; y < Ty+hsub; y++) {
        memcpy(Isub+(y-Ty)*wsub, I+Tx+y*w, wsub);
    }
}



//dray line in image, with gray value 'grey'
int drawgreyline(int x1, int y1, int x2, int y2, unsigned char grey, 
                 unsigned char *out, int width, int height)
{
    int x, y;
    float lambda, d12;

    // distance between corner points
    d12=(float) sqrt((float) ((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)));

    if (d12 == 0.0) return 0;

    for (lambda=0.0f; lambda <= 1.0f; lambda+=(1/d12)) {
        // round float values to closest integers
        x=x1 + (int) roundf(lambda * (float) (x2-x1));
        y=y1 + (int) roundf(lambda * (float) (y2-y1));
        if ((x < width) && (x >= 0) && (y < height) && (y >= 0))
           out[x+y*width]=grey;  //assign gray level to pixel
    }

    return 1;
}

//Draw rectangle in image, with gray value 'grey'
//Top-left corner: (x1, y1)
//Bottom-right corner: (x2, y2)
void drawRectangleGrey(unsigned char *out, int width, int height, 
                       int x1, int y1, int x2, int y2, unsigned char grey)
{
    drawgreyline(x1, y1, x2, y1, grey, out, width, height);
    drawgreyline(x1, y1, x1, y2, grey, out, width, height);
    drawgreyline(x2, y1, x2, y2, grey, out, width, height);
    drawgreyline(x1, y2, x2, y2, grey, out, width, height);
}


//This function computes the intersection between two sub-windows,
//and it is used in Algorithm 6 of the associated IPOL publication

//compute percentage of intersecion area between two rectangles
float check_superposition_rectangles(int x1_a, int y1_a, int x2_a, int y2_a,
                                     int x1_b, int y1_b, int x2_b, int y2_b)
{
    float x1a, y1a, x2a, y2a, x1b, y1b, x2b, y2b;
    float wa, ha, wb, hb, xCa, yCa, xCb, yCb, dx, dy;
    //intersection rectangle: A(top,left)-B(bottom,right)
    float area1, area2, area12, xA, yA, xB, yB; 
    float prccommon;

    x1a=(float) x1_a;
    y1a=(float) y1_a;
    x2a=(float) x2_a;
    y2a=(float) y2_a;
    x1b=(float) x1_b;
    y1b=(float) y1_b;
    x2b=(float) x2_b;
    y2b=(float) y2_b;
    wa=x2a-x1a;
    ha=y2a-y1a;
    wb=x2b-x1b;
    hb=y2b-y1b;
    area1=wa*ha;
    area2=wb*hb;
    xA=xB=0;
    yA=yB=0;
    if ((x1b >= x1a) && (x1b <= x2a)) {
        xA=x1b;
        if (x2b <= x2a) xB=x2b;
        else xB=x2a;
    } else {
        if (x1b < x1a) {
            if ((x2b >= x1a) && (x2b <= x2a)) {
                xA=x1a;
                xB=x2b;
            }
            if (x2b > x2a) {
                xA=x1a;
                xB=x2a;
            }
        }
    }
    if ((y1b >= y1a) && (y1b <= y2a)) {
        yA=y1b;
        if (y2b <= y2a) yB=y2b;
        else yB=y2a;
    } else {
        if (y1b < y1a) {
            if ((y2b >= y1a) && (y2b <= y2a)) {
                yA=y1a;
                yB=y2b;
            }
            if (y2b > y2a) {
                yA=y1a;
                yB=y2a;
            }
        }
    }

    area12=(xB-xA)*(yB-yA);

    //maximum % of common area
    prccommon=area12/area1;
    if (area12/area2 > prccommon) prccommon=area12/area2;
    
    return prccommon;
    
}



